/*----------------------------------------------------------------------------
 * centipedescriptv2.cpp
 *
 * Author: Brian Jimenez Garcia
 * Date: 27-MAR-07
 * Release: 1.0
 * Purpose: Generates a Webots .wbt file according to the centipede.conf file
 * and the configuration chosen with the GUI. This application has been 
 * developed with Qt libraries.
 *----------------------------------------------------------------------------
 */

#include "centipedescriptv2.h"

/* Qt Includes */
#include <qvariant.h>
#include <qpushbutton.h>
#include <qgroupbox.h>
#include <qlistbox.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qlayout.h>
#include <qtooltip.h>
#include <qwhatsthis.h>
#include <qaction.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qtoolbar.h>
#include <qimage.h>
#include <qpixmap.h>
#include <qmessagebox.h>
#include <qfile.h>
#include <qfiledialog.h>
#include <qstatusbar.h>
/* C++ STL Includes */
#include <vector>
#include <iostream>
#include <fstream>
#include <iomanip>
/* User Includes */
#include "image_res.h"
#include "defines.h"

/********************************************/
#define CONFSCRIPT "./centipedescript.conf"
#define CONFFILE "./centipede.conf"
/********************************************/

// double variables to read from configuation file
double femur_length, femur_radius, femur_angle_vert,
    femur_angle_horiz, tibia_length, tibia_radius, tibia_angle,
    foot_radius, foot_height, foot_trans_forwards, foot_trans_inwards,
    ankle_angle, servo_angle_vert, servo_angle_horiz, coulomb_friction,
    force_dependent_slip, body_density, leg_density, bounce, max_force,
    aquarium_width, aquarium_length, water_level, bottom_level, shore_level,
    slope_angle, start_x_pos, start_y_pos, start_z_pos, servo_horiz_stop,
    max_leg_velocity;

// int variables to read from configuation file
int segment_number, attach_point_front, attach_point_back, snake, physics, fly;

// local variables
double leg_servo_x_coeff, leg_servo_y_coeff, leg_servo_z_coeff;

/* structures for storing the centipede's limb parameters */
struct s_part {
  /* translation coefficients */
  double transl_x, transl_y, transl_z, rot_angle;

  /* rotation axis vector */
  double rot_x, rot_y, rot_z;
};

struct leg {
  struct s_part femur, knee, servo_tibia, tibia, foot;
};

// centipede legs
struct leg *legs;

static int indent_level = 0;

static int head = 0;
static int rigidlegs = 0;
static int nlegs = 0;

/************************************************************************************
 *
 * 			   QT GRAPHICAL INTERFACE CODE
 *
 ***********************************************************************************/

/*
 *  Constructs a CentipedeScript as a child of 'parent', with the
 *  name 'name' and widget flags set to 'f'.
 *
 */
CentipedeScript::CentipedeScript( QWidget* parent, const char* name, WFlags fl )
    : QMainWindow( parent, name, fl ),
      image1( (const char **) image1_data )
{
	readConfScriptFile(CONFSCRIPT);
	
    (void)statusBar();
    QImage img;
    img.loadFromData( image0_data, sizeof( image0_data ), "PNG" );
    image0 = img;
    if ( !name )
	setName( "CentipedeScript" );
    setIcon( image0 );
    setCentralWidget( new QWidget( this, "qt_central_widget" ) );
    CentipedeScriptLayout = new QVBoxLayout( centralWidget(), 11, 6, "CentipedeScriptLayout"); 

    groupBox2 = new QGroupBox( centralWidget(), "groupBox2" );
    groupBox2->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)3, (QSizePolicy::SizeType)3, 0, 0, groupBox2->sizePolicy().hasHeightForWidth() ) );

    listModules = new QListBox( groupBox2, "listModules" );
    listModules->setGeometry( QRect( 30, 30, 260, 90 ) );
    listModules->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)3, (QSizePolicy::SizeType)3, 0, 0, listModules->sizePolicy().hasHeightForWidth() ) );
    listModules->setMinimumSize( QSize( 260, 90 ) );

    buttonAdd = new QPushButton( groupBox2, "buttonAdd" );
    buttonAdd->setGeometry( QRect( 30, 130, 220, 21 ) );
    buttonAdd->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)3, (QSizePolicy::SizeType)0, 0, 0, buttonAdd->sizePolicy().hasHeightForWidth() ) );
    buttonAdd->setToggleButton( FALSE );

    pixImage = new QLabel( groupBox2, "pixImage" );
    pixImage->setGeometry( QRect( 320, 30, 121, 121 ) );
    pixImage->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)3, (QSizePolicy::SizeType)3, 0, 0, pixImage->sizePolicy().hasHeightForWidth() ) );
    pixImage->setPaletteBackgroundColor( QColor( 255, 255, 255 ) );
    pixImage->setCursor( QCursor( 2 ) );
    pixImage->setLineWidth( 1 );
    pixImage->setScaledContents( TRUE );
	pixImage->setPixmap(QPixmap(cf[0].image_path,"PNG"));
    CentipedeScriptLayout->addWidget( groupBox2 );
    spacer4 = new QSpacerItem( 420, 60, QSizePolicy::Minimum, QSizePolicy::MinimumExpanding );
    CentipedeScriptLayout->addItem( spacer4 );

    groupBox3 = new QGroupBox( centralWidget(), "groupBox3" );
    groupBox3->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)3, (QSizePolicy::SizeType)0, 0, 0, groupBox3->sizePolicy().hasHeightForWidth() ) );

    lineConfiguration = new QLineEdit( groupBox3, "lineConfiguration" );
    lineConfiguration->setGeometry( QRect( 30, 30, 340, 24 ) );
    lineConfiguration->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)3, (QSizePolicy::SizeType)0, 0, 0, lineConfiguration->sizePolicy().hasHeightForWidth() ) );
    lineConfiguration->setCursor( QCursor( 4 ) );
	lineConfiguration->setEnabled(false);

    buttonDel = new QPushButton( groupBox3, "buttonDel" );
    buttonDel->setGeometry( QRect( 380, 30, 21, 21 ) );
    buttonDel->setPixmap( image1 );
    CentipedeScriptLayout->addWidget( groupBox3 );
    spacer6 = new QSpacerItem( 421, 40, QSizePolicy::Minimum, QSizePolicy::MinimumExpanding );
    CentipedeScriptLayout->addItem( spacer6 );

    groupBox3_2 = new QGroupBox( centralWidget(), "groupBox3_2" );
    groupBox3_2->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)3, (QSizePolicy::SizeType)0, 0, 0, groupBox3_2->sizePolicy().hasHeightForWidth() ) );

    comboScenario = new QComboBox( FALSE, groupBox3_2, "comboScenario" );
    comboScenario->setGeometry( QRect( 30, 19, 340, 21 ) );
    CentipedeScriptLayout->addWidget( groupBox3_2 );
    spacer6_2 = new QSpacerItem( 421, 40, QSizePolicy::Minimum, QSizePolicy::MinimumExpanding );
    CentipedeScriptLayout->addItem( spacer6_2 );

    layout3 = new QHBoxLayout( 0, 0, 6, "layout3"); 

    buttonExit = new QPushButton( centralWidget(), "buttonExit" );
    buttonExit->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)3, (QSizePolicy::SizeType)0, 0, 0, buttonExit->sizePolicy().hasHeightForWidth() ) );
    layout3->addWidget( buttonExit );
    spacer5 = new QSpacerItem( 160, 31, QSizePolicy::MinimumExpanding, QSizePolicy::Minimum );
    layout3->addItem( spacer5 );

    buttonGenerate = new QPushButton( centralWidget(), "buttonGenerate" );
    buttonGenerate->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)3, (QSizePolicy::SizeType)0, 0, 0, buttonGenerate->sizePolicy().hasHeightForWidth() ) );
    layout3->addWidget( buttonGenerate );
    CentipedeScriptLayout->addLayout( layout3 );

    // toolbars
    languageChange();
    resize( QSize(445, 435).expandedTo(minimumSizeHint()) );
    clearWState( WState_Polished );

    // signals and slots connections
    connect( buttonExit, SIGNAL( clicked() ), this, SLOT( close() ) );
	connect( buttonGenerate, SIGNAL( clicked() ), this, SLOT( save() ) );
	connect( buttonAdd, SIGNAL( clicked() ), this, SLOT( addModule() ) );
	connect( buttonDel, SIGNAL( clicked() ), this, SLOT( delModule() ) );
	connect( listModules, SIGNAL( highlighted(int) ), this, SLOT( changedModule(int) ) );
}

/*
 *  Destroys the object and frees any allocated resources
 */
CentipedeScript::~CentipedeScript()
{
    // no need to delete child widgets, Qt does it all for us
}

/*
 *  Sets the strings of the subwidgets using the current
 *  language.
 */
void CentipedeScript::languageChange()
{
    setCaption( tr( "CentipedeScript" ) );
    groupBox2->setTitle( tr( "Robot Modules" ) );
    buttonAdd->setText( tr( "&Add" ) );
    buttonAdd->setAccel( QKeySequence( tr( "Alt+A" ) ) );
    QToolTip::add( buttonAdd, tr( "Add an active module to the actual configuration" ) );
    QWhatsThis::add( buttonAdd, QString::null );
    listModules->clear();
    for(uint i=0; i<cf.size(); ++i)
    	listModules->insertItem( tr( cf[i].name ) );
    listModules->setCurrentItem( 0 );
    QToolTip::add( listModules, tr( "List of available modules" ) );
    QWhatsThis::add( listModules, QString::null );
    QToolTip::add( pixImage, tr( "Representation of the active selected module" ) );
    QWhatsThis::add( pixImage, QString::null );
    groupBox3->setTitle( tr( "Actual Configuration" ) );
    QToolTip::add( lineConfiguration, tr( "Actual configuration" ) );
    QWhatsThis::add( lineConfiguration, QString::null );
    buttonDel->setText( QString::null );
    QToolTip::add( buttonDel, tr( "Deletes the last added module" ) );
	groupBox3_2->setTitle( tr( "Scenario" ) );
    buttonExit->setText( tr( "&Exit" ) );
    buttonExit->setAccel( QKeySequence( tr( "Alt+E" ) ) );
    QToolTip::add( buttonExit, tr( "Exits the application" ) );
    QWhatsThis::add( buttonExit, QString::null );
    buttonGenerate->setText( tr( "&Generate" ) );
    buttonGenerate->setAccel( QKeySequence( tr( "Alt+G" ) ) );
    QToolTip::add( buttonGenerate, tr( "Generates a new Webots world according to the actual configuration" ) );
	comboScenario->insertItem("Aquarium without Water",0);
	comboScenario->insertItem("Aquarium with Water",1);
	comboScenario->insertItem("Infinite Ground",2);
	comboScenario->insertItem("Infinite Water",3);
	comboScenario->insertItem("Grid World",4);
}

/*
 * Captures the close event
 *
 */
void CentipedeScript::closeEvent( QCloseEvent* ce )
{
    switch( QMessageBox::information( this, "CentipedeScript",
                                      "Exit the application?",
                                      "Yes", "No",
                                      0, 1 ) ) {
    case 0:
		ce->accept();
        break;
    case 1:
        ce->ignore();
        break;
    }
}

/*
 * Saves and generates the actual configuration of the robot
 * in case that the filename f is not empty, elsewhere it calls
 * saveAs()
 */
void CentipedeScript::save()
{
    if ( filename.isEmpty() ) {
        saveAs();
        return;
    }

    QFile f( filename );
    if ( !f.open( IO_WriteOnly ) ) {
        statusBar()->message( QString("Could not write to %1").arg(filename),
                              2000 );
        return;
    }

    QTextStream t( &f );
    generate(&t);
    f.close();

    statusBar()->message( QString( "File %1 saved" ).arg( filename ), 2000 );
}

/*
 * Get the filename f from a Dialog and then calls save()
 *
 */
void CentipedeScript::saveAs()
{
    QString fn = QFileDialog::getSaveFileName( QString::null, ".wbt",
                                               this );
    if ( !fn.isEmpty() ) {
        filename = fn+".wbt" ;
        save();
    } else {
        statusBar()->message( "Generating aborted", 2000 );
    }
}

/*
 * Adds a module in the lineConfiguration LineEdit widget.
 */
void CentipedeScript::addModule()
{
	// Only add a head one time
	if(cf[listModules->currentItem()].abbr=="C" || cf[listModules->currentItem()].abbr=="D") {
		if(!head) {
			head = 1;
			lineConfiguration->setText(lineConfiguration->text() + "(" + cf[listModules->currentItem()].abbr + ")");
			robotConf.push_back(listModules->currentItem());
		}
	} else {
		lineConfiguration->setText(lineConfiguration->text() + "(" + cf[listModules->currentItem()].abbr + ")");
		robotConf.push_back(listModules->currentItem());
	}
}

/*
 * Removes a module in the lineConfiguration LineEdit widget.
 */
void CentipedeScript::delModule()
{
	int i = lineConfiguration->text().findRev( '(' );
	lineConfiguration->setText(lineConfiguration->text().remove(i,lineConfiguration->text().length()-i));
	robotConf.pop_back();
	if(robotConf.size()==0) head=0;
}

/*
 * Changes the image according to the module selected
 */
void CentipedeScript::changedModule(int index)
{
	pixImage->setPixmap(QPixmap(cf[index].image_path,"PNG"));
}

/*
 * Get the configuration file through a dialog box
 *
 */
QString CentipedeScript::getConfFile()
{
	QString fn = QFileDialog::getOpenFileName( "centipedescript.conf", QString::null, this );
    return fn;
}


/*
 * Reads configuration file <s>
 *
 */
void CentipedeScript::readConfScriptFile( QString s )
{
		
	char buffer[256];
	
	// Open the file
	std::ifstream file(s, std::ios::in);
	
	if(!file.good()) {
	    switch( QMessageBox::information( this, "CentipedeScript",
                                      "Configuration file not found",
                                      "Open it manually", "Exit",
                                      0, 1 ) ) {
    	case 0:
			file.open (getConfFile(), std::ifstream::in);
			if(!file.good()) exit(FILE_NO_EXISTS);
        	break;
    	case 1:
        	exit(FILE_NO_EXISTS);
        	break;
    	}
	}
	
	// Read the file
	while (file.good()) {
	
		file.getline(buffer,256);

		if(buffer[0] == '<') {
			std::string line(buffer);
			int nump = 0;
			int p[2]; for(int j=0;j<2;++j) p[j]=0;
			int g = 0;
			ConfFile info;
						
			for(uint i=0;i<line.size();++i) {
				if(line[i]==';') {
					p[nump++]=i;
					// Correct file structure
					if(nump > 2) exit(FILE_ERROR);
				}
				// Reading until the first dash
				if(line[i]=='>') g=i;
			}
			
			if(nump != 2) exit(FILE_ERROR);

			info.name = QString(line.substr(1,p[0]-1)).simplifyWhiteSpace();	
			info.abbr = QString(line.substr(p[0]+1,p[1]-p[0]-1)).simplifyWhiteSpace();
			info.image_path = QString(line.substr(p[1]+1,g-p[1]-1)).simplifyWhiteSpace();
			cf.push_back(info);
		}
	}
	
	file.close();	
}

/*
 * retrieve variables from configuration file (added by Yvan)
 */
void CentipedeScript::readConfFile ( QString s )
{
  /* open config file */
  FILE *config_file = fopen(CONFFILE, "r");
  if (!config_file) {
    fprintf(stderr, "Error: cannot open file: %s\n", CONFFILE);
    exit(FAILURE);
  }

  struct {
    char *name;
    void *var;
    char type;
  } P[] = {
    {
    /*"segment_number", &segment_number, 'd'}, {
    "attach_point_front", &attach_point_front, 'd'}, {*/
    "fly", &fly, 'd'}, {
    "physics", &physics, 'd'}, {
    "max_force", &max_force, 'f'}, {
    "femur_length", &femur_length, 'f'}, {
    "femur_radius", &femur_radius, 'f'}, {
    "femur_angle_vert", &femur_angle_vert, 'f'}, {
    "femur_angle_horiz", &femur_angle_horiz, 'f'}, {
    "tibia_length", &tibia_length, 'f'}, {
    "tibia_radius", &tibia_radius, 'f'}, {
    "tibia_angle", &tibia_angle, 'f'}, {
    "foot_radius", &foot_radius, 'f'}, {
    "foot_height", &foot_height, 'f'}, {
    "foot_trans_forwards", &foot_trans_forwards, 'f'}, {
    "foot_trans_inwards", &foot_trans_inwards, 'f'}, {
    "ankle_angle", &ankle_angle, 'f'}, {
    "servo_angle_vert", &servo_angle_vert, 'f'}, {
    "servo_angle_horiz", &servo_angle_horiz, 'f'}, {
    "coulomb_friction", &coulomb_friction, 'f'}, {
    "force_dependent_slip", &force_dependent_slip, 'f'}, {
    "body_density", &body_density, 'f'}, {
    "leg_density", &leg_density, 'f'}, {
    "bounce", &bounce, 'f'}, {
    "aquarium_width", &aquarium_width, 'f'}, {
    "aquarium_length", &aquarium_length, 'f'}, {
    "water_level", &water_level, 'f'}, {
    "bottom_level", &bottom_level, 'f'}, {
    "shore_level", &shore_level, 'f'}, {
    "slope_angle", &slope_angle, 'f'}, {
    "start_x_pos", &start_x_pos, 'f'}, {
    "start_y_pos", &start_y_pos, 'f'}, {
    "start_z_pos", &start_z_pos, 'f'}, {
    "servo_horiz_stop", &servo_horiz_stop, 'f'}, {
    "max_leg_velocity", &max_leg_velocity, 'f'}, {
    NULL}
  };

  /* start reading */
 char line[128];
  while (fgets(line, 128, config_file)) {
    char value[80], variable[80];
    if (line[0] != '#' && sscanf(line, "%s%s", variable, value) == 2) {
      int i;
      for (i = 0; P[i].name; i++) {
        if (!strcmp(variable, P[i].name)) {
          switch (P[i].type) {
          case 'd':
            *((int *) (P[i].var)) = atoi(value);
            break;
          case 'f':
            *((double *) (P[i].var)) = atof(value);
            break;
          }
          break;
        }
      }
    }
  }
  
  fclose(config_file);
}


/************************************************************************************
 *
 * 				ROBOT CONSTRUCTION
 *
 ***********************************************************************************/

/*
 * Generates the Webots world file according to the configuration selected
 */
void CentipedeScript::generate( QTextStream *t)
{
	rigidlegs = 0;
	readConfFile("");

	femur_angle_vert = DEG_TO_RAD(femur_angle_vert);
	femur_angle_horiz = DEG_TO_RAD(femur_angle_horiz);
	tibia_angle = DEG_TO_RAD(tibia_angle);
	servo_angle_vert = DEG_TO_RAD(servo_angle_vert);
	servo_angle_horiz = DEG_TO_RAD(servo_angle_horiz);
	ankle_angle = DEG_TO_RAD(ankle_angle);
	slope_angle = DEG_TO_RAD(slope_angle);
	servo_horiz_stop = DEG_TO_RAD(servo_horiz_stop);

	double x_axis[3], femur_axis[3], rot_axis[3];		

	// Capture the number of legs
	nlegs = 0;
	for(uint i=0; i<robotConf.size();++i)
	{
		// Add any new module with legs here
		if(isLeg(cf[robotConf[i]].abbr)) nlegs+=2;
		if(cf[robotConf[i]].abbr=="E") rigidlegs=1;
	}
	legs = new struct leg[nlegs];

  	// Compute leg parameters
	if(nlegs > 0) {
	    	/* servo axis orientation */
	    	leg_servo_x_coeff = sin(PI_2 - servo_angle_vert) * cos(servo_angle_horiz);
		leg_servo_z_coeff = sin(PI_2 - servo_angle_vert) * sin(servo_angle_horiz);
		leg_servo_y_coeff = cos(PI_2 - servo_angle_vert);

	    	/* compute translation of the femur's center of gravity */
	    	legs[0].femur.transl_x = -femur_length / 2;

	    	/* compute the rotation axis orientation. used spherical coordinates */

	    	/* -X axis */
	    	x_axis[X] = 1;
	    	x_axis[Y] = x_axis[Z] = 0;

	    	/* femur orientation */
	    	femur_axis[X] = sin(PI_2 - femur_angle_vert) * cos(femur_angle_horiz);
	    	femur_axis[Y] = cos(PI_2 - femur_angle_vert);
	    	femur_axis[Z] = sin(PI_2 - femur_angle_vert) * sin(femur_angle_horiz);

	    	/* get a vector normal to X axis and femur position */
	    	VECTOR_PRODUCT(x_axis, femur_axis, rot_axis);

	    	legs[0].femur.rot_x = rot_axis[X];
	    	legs[0].femur.rot_y = rot_axis[Y];
	    	legs[0].femur.rot_z = rot_axis[Z];

	    	/* normalize the vector */
	    	double norm = NORM3(legs[0].femur.rot_x, legs[0].femur.rot_y, legs[0].femur.rot_z);

	    	legs[0].femur.rot_x /= norm;
	    	legs[0].femur.rot_y /= norm;
	    	legs[0].femur.rot_z /= norm;

	    	/* compute the angle between the two vectors. */
	    	legs[0].femur.rot_angle = acos(SCALAR_PRODUCT(x_axis, femur_axis));
	    	legs[0].knee.transl_x = -femur_length;
	    	legs[0].servo_tibia.transl_x = legs[0].knee.transl_x;
	    	legs[0].servo_tibia.rot_angle = tibia_angle;
	    	legs[0].tibia.transl_x = -tibia_length / 2;
	    	legs[0].foot.transl_x = 2 * legs[0].tibia.transl_x;
	    	legs[0].foot.transl_y = -foot_trans_inwards;
    		legs[0].foot.transl_z = -foot_trans_forwards;

		for(int l=1; l<nlegs; ++l) {
	    		legs[l].femur.transl_x = (l%2==1)? femur_length / 2 : legs[0].femur.transl_x;
	    		legs[l].femur.rot_x = -legs[0].femur.rot_x;
	    		legs[l].femur.rot_y = legs[0].femur.rot_y;
	    		legs[l].femur.rot_z = legs[0].femur.rot_z;
	    		legs[l].femur.rot_angle = -legs[0].femur.rot_angle;
	    		legs[l].knee.transl_x = (l%2==1)? femur_length : legs[0].knee.transl_x;
	    		legs[l].servo_tibia.transl_x = legs[l].knee.transl_x;
	    		legs[l].servo_tibia.rot_angle = tibia_angle;
	    		legs[l].tibia.transl_x = (l%2==1)? tibia_length / 2 : legs[0].tibia.transl_x;
	    		legs[l].foot.transl_x = 2 * legs[l].tibia.transl_x;
	    		legs[l].foot.transl_y = foot_trans_inwards;
	    		legs[l].foot.transl_z = (l%2==1)? foot_trans_forwards : -foot_trans_forwards;
		}
	}
	
	// Write the Webots World
	writeWorld(t);
	if(fly) writeFly(t);
	writeHeader(t);
	if(cf[robotConf[0]].abbr=="D") writeCameras(t);

  	indent_level = 4;

	int first_legs = 0;
	int l = 0;
	double ztranslation = 0;
	// According to the robot segments, compute the outfile file
	for(uint i=1; i< robotConf.size(); ++i)
	{
		*t << "DEF SERVO_" << i << " Servo {" << "\n";
		indent_level += 2;
		if(isLeg(cf[robotConf[i]].abbr)) {
			ztranslation = BIG_SEGMENT_LENGTH;
			if(!isLeg(cf[robotConf[i-1]].abbr)) ztranslation += BIG_SEGMENT_WIDTH_HALF;
		} else {
			ztranslation = SEGMENT_LENGTH;
			if(isLeg(cf[robotConf[i-1]].abbr)) ztranslation -= BIG_SEGMENT_WIDTH_HALF;
		}

		*t << "translation 0 0 " << ztranslation << "\n";
    		*t << "rotation 0 1 0 0" << "\n";
	    	*t << "children [" << "\n";
    		indent_level += 2;
		
		// Normal segment with legs
		if(isLeg(cf[robotConf[i]].abbr)) {
			// First module with legs?				
			if(!first_legs) {
				first_legs = 1;
				writeDefBigSegment(t);
				if(cf[robotConf[i]].abbr=="B") {
					writeLeg(t,++l);
					writeLeg(t,++l);
				} else {
					writeRigidLeg(t,++l);
					writeRigidLeg(t,++l);
				}
			} else {
				*t << "USE BIG_SEGMENT" << "\n";
				if(cf[robotConf[i]].abbr=="B") {
					writeLeg(t,++l);
					writeLeg(t,++l);
				} else {
					writeRigidLeg(t,++l);
					writeRigidLeg(t,++l);
				}
			}
		}

		// Normal segment without legs
		if(!isLeg(cf[robotConf[i]].abbr)) {
			*t << "USE SMALL_SEGMENT" << "\n";
		}
	}
	
	for (int i = robotConf.size(); i > 0; i--) 
	{
    		indent_level -= 2;
    		*t << "]" << "\n";
    		// the head segment is not a Servo but a CustomRobot
    		if (i != 1) *t << "name \"servo_" << i-1 << "\"" << "\n";

    		// handle two different types of bounding objects
    		if (isLeg(cf[robotConf[i-1]].abbr)) *t << "boundingObject USE BIG_SEGMENT" << "\n";
    		else *t << "boundingObject USE SMALL_SEGMENT" << "\n";				//Modify if new modules
		
    		*t << "physics Physics {" << "\n";
    		indent_level += 2;
    		*t << "density " << body_density << "\n";
    		*t << "bounce " << bounce << "\n";
    		*t << "coulombFriction " << coulomb_friction << "\n";
    		*t << "forceDependentSlip " << force_dependent_slip << "\n";
		
    		// since Webots5 it is possible to set the center of mass of bodies
		// added by yvan.bourquin@epfl.ch
		double mass_trans_z;
		double mass_trans_y = -SEGMENT_HEIGHT * 0.25;
		if (isLeg(cf[robotConf[i-1]].abbr))
		mass_trans_z = (BIG_SEGMENT_LENGTH - BIG_SEGMENT_WIDTH_HALF) / 2.0;
		else
		mass_trans_z = (SEGMENT_LENGTH - SEGMENT_WIDTH_HALF) / 2.0;

		*t << "centerOfMass 0 "<< mass_trans_y << " " << mass_trans_z << "\n";

		indent_level -= 2;
		*t << "}" << "\n";
		*t << "joint Joint {" << "\n";
		*t << "}" << "\n";

		if (i > 1) {
			// body segment
			*t << "maxForce " << max_force << "\n";
			
			if (isLeg(cf[robotConf[i-1]].abbr)) {
				// the horizontal servos of legged elements must
				// be blocked because they don't exist on the actual robot
				*t << "minPosition -0.0001" << "\n";
				*t << "maxPosition 0.0001" << "\n";
			}
			else {
				*t << "minPosition " << -servo_horiz_stop << "\n";
				*t << "maxPosition " << servo_horiz_stop << "\n";
			}
		}
		else
			// the head segment is the robot
			//*t << "controller \"centipede" << (rigidlegs ? "_rigid" : "" ) << nlegs/2 << "\"" << "\n";
			*t << "controller \"void\"\n";
			
		indent_level -= 2;
		*t << "}" << "\n";

	}    	

	// Scenario configuration
	switch(comboScenario->currentItem())
	{
		case 0:
			writeGround(t);
			writeAquarium(t);
			break;
		case 1:
			writeGround(t);
			writeAquarium(t);
			writeWater(t);
			break;
		case 2:
			writeInfiniteGround(t);
			break;
		case 3:
			writeInfiniteWater(t);
			break;
		case 4:
			writeGridGround(t);
			break;
	}
}


/*
 * Writes the Centipede Robot Header
 */
void CentipedeScript::writeHeader( QTextStream *t )
{
  *t << "DEF CENTIPEDE CustomRobot {" << "\n";
  *t << "  translation " << start_x_pos << " " << start_y_pos << " " << start_z_pos << "\n";
  *t << "  rotation 0 1 0 1.5808" << "\n";
  *t << "  children [" << "\n";

  indent_level += 4;
  writeDefSmallSegment(t);
  indent_level -= 4;

}

/*
 * Writes the robot cameras on the head. Adapted from Yvan.
 */
void CentipedeScript::writeCameras( QTextStream *t )  {
  indent_level += 4;
  *t << "DEF RIGHT_CAMERA Camera {" << "\n";
  *t << "  translation -0.02 0.03 0" << "\n";
  *t << "  rotation 0 1 0 0.5236" << "\n";
  *t << "  children [" << "\n";
  *t << "    DEF EYE_SHAPE Shape {" << "\n";
  *t << "      appearance DEF EYE_APPEARANCE Appearance {" << "\n";
  *t << "        material Material {" << "\n";
  *t << "          diffuseColor 0 0 0" << "\n";
  *t << "        }" << "\n";
  *t << "      }" << "\n";
  *t << "      geometry Sphere {" << "\n";
  *t << "        radius 0.005" << "\n";
  *t << "      }" << "\n";
  *t << "    }" << "\n";
  *t << "  ]" << "\n";
  *t << "  name \"right_camera\"" << "\n";
  *t << "  fieldOfView 2.0944" << "\n";
  *t << "  width 128" << "\n";
  *t << "  height 64" << "\n";
  *t << "}" << "\n";
  *t << "DEF LEFT_CAMERA Camera {" << "\n";
  *t << "  translation 0.02 0.03 0" << "\n";
  *t << "  rotation 0 1 0 -0.5236" << "\n";
  *t << "  children [" << "\n";
  *t << "    USE EYE_SHAPE" << "\n";
  *t << "  ]" << "\n";
  *t << "  name \"left_camera\"" << "\n";
  *t << "  fieldOfView 2.0944" << "\n";
  *t << "  width 128" << "\n";
  *t << "  height 64" << "\n";
  *t << "}" << "\n";
  indent_level -= 4;
}

/*
 * Writes a big segment of the model. Adapted from Yvan
 */
void CentipedeScript::writeDefBigSegment( QTextStream *t )
{
  double z_shift = BIG_SEGMENT_LENGTH - BIG_SEGMENT_WIDTH_HALF;

  *t << "DEF BIG_SEGMENT Group {" << "\n";
  *t << "  children [" << "\n";
  *t << "    Transform {" << "\n";
  *t << "      children [" << "\n";
  *t << "        Shape {" << "\n";
  *t << "          appearance USE SALAMANDER_COLOR" << "\n";
  *t << "          geometry Cylinder {" << "\n";
  *t << "            height " << SEGMENT_HEIGHT << "\n";
  *t << "            radius " << BIG_SEGMENT_WIDTH / 2.0 << "\n";
  *t << "          }" << "\n";
  *t << "        }" << "\n";
  *t << "      ]" << "\n";
  *t << "    }" << "\n";
  *t << "    Transform {" << "\n";
  *t << "      translation 0 0 " << z_shift / 2.0 << "\n";
  *t << "      children [" << "\n";
  *t << "        Shape {" << "\n";
  *t << "          appearance USE SALAMANDER_COLOR" << "\n";
  *t << "          geometry Box {" << "\n";
  *t << "            size " << BIG_SEGMENT_WIDTH << " " << SEGMENT_HEIGHT << " " << z_shift << "\n";
  *t << "          }" << "\n";
  *t << "        }" << "\n";
  *t << "      ]" << "\n";
  *t << "    }" << "\n";
  *t << "  ]" << "\n";
  *t << "}" << "\n";
}

/*
 * Writes a small segment of the model. Adapted from Yvan.
 */
void CentipedeScript::writeDefSmallSegment( QTextStream *t ) {
  
  double z_shift = SEGMENT_LENGTH - SEGMENT_WIDTH_HALF;

  *t << "DEF SMALL_SEGMENT Group {" << "\n";
  *t << "  children [" << "\n";
  *t << "    Transform {" << "\n";
  *t << "      children [" << "\n";
  *t << "        Shape {" << "\n";
  *t << "        appearance DEF SALAMANDER_COLOR Appearance {" << "\n";
  *t << "            material Material {" << "\n";
  *t << "              emissiveColor 1 0.8 0.3" << "\n";
  *t << "            }" << "\n";
  *t << "          }" << "\n";
  *t << "          geometry Cylinder {" << "\n";
  *t << "            height " << SEGMENT_HEIGHT << "\n";
  *t << "            radius " << SEGMENT_WIDTH / 2.0  << "\n";
  *t << "          }" << "\n";
  *t << "        }" << "\n";
  *t << "      ]" << "\n";
  *t << "    }" << "\n";
  *t << "    Transform {" << "\n";
  *t << "      translation 0 0 " << z_shift / 2.0 << "\n";
  *t << "      children [" << "\n";
  *t << "        Shape {" << "\n";
  *t << "          appearance USE SALAMANDER_COLOR" << "\n";
  *t << "          geometry Box {" << "\n";
  *t << "            size " << SEGMENT_WIDTH << " " << SEGMENT_HEIGHT << " " << z_shift << "\n";
  *t << "          }" << "\n";
  *t << "        }" << "\n";
  *t << "      ]" << "\n";
  *t << "    }" << "\n";
  *t << "  ]" << "\n";
  *t << "}" << "\n";
}

/*
 * Writes the model's legs. Adapted from Yvan.
 */
void CentipedeScript::writeLeg( QTextStream *t, int leg_number ) {
  // odd number means left side
  int left = leg_number % 2 == 1;
  int leg_index = leg_number - 1;

  *t << "DEF SERVO_LEG_" << leg_number << " Servo {" << "\n";
  *t << "  translation "<< (left ? -BIG_SEGMENT_WIDTH_HALF : BIG_SEGMENT_WIDTH_HALF) << " 0 " << LEGS_BACKWARDS_SHIFT << "\n";
  *t << "  rotation 1 0 0 " << (left ? 0.0 : PI) << "\n";
  *t << "  children [" << "\n";
  *t << "    DEF GROUP_FEMUR_"<< leg_number <<" Group {" << "\n";
  *t << "     children [" << "\n";
  *t << "        DEF TRANSFORM_HIP_"<< leg_number <<" Transform {" << "\n";
  *t << "          translation "<< (left ? -0.004 : 0.004) << " 0 0" << "\n";
  *t << "          rotation 0 0 1 "<< PI_2 << "\n";
  *t << "          children [" << "\n";

  if (leg_number == 1) {
    *t << "            DEF SHAPE_HIP Shape {" << "\n";
    *t << "              appearance DEF LEG_COLOR Appearance {" << "\n";
    *t << "                material Material {" << "\n";
    *t << "                  emissiveColor 0 0 0" << "\n";
    *t << "                }" << "\n";
    *t << "              }" << "\n";
    *t << "              geometry Cylinder {" << "\n";
    *t << "                height " << HIP_LENGTH << "\n";
    *t << "                radius " << HIP_RADIUS << "\n";
    *t << "              }" << "\n";
    *t << "            }" << "\n";
  } else
    *t << "            USE SHAPE_HIP" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "        DEF TRANSFORM_FEMUR_"<< leg_number <<" Transform {" << "\n";
  *t << "          translation " << legs[leg_index].femur.transl_x + HIP_LENGTH / (left ? -2.0 : 2.0) <<" 0 0" << "\n";
  *t << "          rotation 0 0 1 " << PI_2  << "\n";
  *t << "          children [" << "\n";
  *t << "            DEF FEMUR Shape {" << "\n";
  *t << "              appearance USE LEG_COLOR" << "\n";
  *t << "              geometry Cylinder {" << "\n";
  *t << "                height " << femur_length - HIP_LENGTH  << "\n";
  *t << "                radius " << femur_radius << "\n";
  *t << "              }" << "\n";
  *t << "            }" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "        DEF TRANSFORM_KNEE_"<< leg_number <<" Transform {" << "\n";
  *t << "          translation " << legs[leg_index].knee.transl_x << " 0 0" << "\n";
  *t << "          children [" << "\n";
  *t << "            DEF KNEE Shape {" << "\n";
  *t << "              appearance USE LEG_COLOR" << "\n";
  *t << "              geometry Sphere {" << "\n";
  *t << "                radius " << femur_radius << "\n";
  *t << "              }" << "\n";
  *t << "            }" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "      ]" << "\n";
  *t << "    }" << "\n";
  //
  *t << "    DEF SERVO_TIBIA_"<< leg_number <<" Servo {" << "\n";
  *t << "      translation " << legs[leg_index].servo_tibia.transl_x << " 0 0" << "\n";
  *t << "      rotation 0 0 1 " << legs[leg_index].servo_tibia.rot_angle << "\n";
  *t << "      children [" << "\n";
  //
  *t << "        DEF GROUP_TIBIA_" << leg_number <<" Group {" << "\n";
  *t << "          children [" << "\n";
  *t << "            DEF TRANSFORM_TIBIA_"<< leg_number << " Transform {" << "\n";
  *t << "              translation " << legs[leg_index].tibia.transl_x << " 0 0" << "\n";
  *t << "              rotation 0 0 1 " << PI_2 << "\n";
  *t << "              children [" << "\n";
  *t << "                DEF TIBIA Shape {" << "\n";
  *t << "                  appearance USE LEG_COLOR" << "\n";
  *t << "                  geometry Cylinder {" << "\n";
  *t << "                    height " << tibia_length << "\n";
  *t << "                    radius " << tibia_radius << "\n";
  *t << "                  }" << "\n";
  *t << "                }" << "\n";
  *t << "              ]" << "\n";
  *t << "            }" << "\n";
  *t << "            DEF TRANSFORM_FOOT_"<< leg_number << " Transform {" << "\n";
  *t << "              translation " << legs[leg_index].foot.transl_x << " " << legs[leg_index].foot.transl_y << " " << legs[leg_index].foot.transl_z << "\n";
  *t << "              rotation 0 0 1 " << -PI_2 + ankle_angle << "\n";
  *t << "              children [" << "\n";
  *t << "                DEF FOOT Shape {" << "\n";
  *t << "                  appearance USE LEG_COLOR" << "\n";
  *t << "                  geometry Cylinder {" << "\n";
  *t << "                    height " << foot_height << "\n";
  *t << "                    radius " << foot_radius << "\n";
  *t << "                  }" << "\n";
  *t << "                }" << "\n";
  *t << "              ]" << "\n";
  *t << "            }" << "\n";
  *t << "          ]" << "\n";
  //
  *t << "        }" << "\n";
  *t << "      ]" << "\n";
  *t << "      name \"servo_tibia_" << leg_number << "\"" << "\n";
  *t << "      boundingObject USE GROUP_TIBIA_" << leg_number << "\n";
  *t << "      physics Physics {" << "\n";
  *t << "        density " << leg_density << "\n";
  *t << "        bounce " << BOUNCE_PARAM << "\n";
  *t << "        coulombFriction " << coulomb_friction << "\n";
  *t << "        forceDependentSlip " << force_dependent_slip << "\n";
  *t << "        centerOfMass " << tibia_length * 2.0 / (left ? -3.0 : 3.0) << " 0 0" << "\n";
  *t << "      }" << "\n";
  *t << "      joint Joint {" << "\n";
  *t << "      }" << "\n";
  *t << "      maxForce " << max_force << "\n";
  *t << "      minPosition -0.0001" << "\n"; // this blocks the tibia servos
  *t << "      maxPosition 0.0001" << "\n";
  //
  *t << "    }" << "\n";
  *t << "  ]" << "\n";
  *t << "  name \"servo_leg_"<< leg_number << "\"" << "\n";
  *t << "  boundingObject USE TRANSFORM_FOOT_" << leg_number << "\n";
  *t << "    physics Physics {" << "\n";
  *t << "      density " << leg_density << "\n";
  *t << "      bounce " << BOUNCE_PARAM << "\n";
  *t << "      coulombFriction " << coulomb_friction << "\n";
  *t << "      forceDependentSlip " << force_dependent_slip << "\n";
  *t << "      centerOfMass " << femur_length / (left ? -2.0 : 2.0) << " 0 0" << "\n";
  *t << "    }" << "\n";
  *t << "  joint Joint {" << "\n";
  *t << "  }" << "\n";
  *t << "  maxVelocity " << max_leg_velocity << "\n";
  *t << "  maxForce " << max_force << "\n";
  *t << "}" << "\n";
}

/*
 * Writes the model's legs with a rigid tibia
 */
void CentipedeScript::writeRigidLeg( QTextStream *t, int leg_number ) {
  // odd number means left side
  int left = leg_number % 2 == 1;
  int leg_index = leg_number - 1;

  *t << "DEF SERVO_LEG_" << leg_number << " Servo {" << "\n";
  *t << "  translation "<< (left ? -BIG_SEGMENT_WIDTH_HALF : BIG_SEGMENT_WIDTH_HALF) << " 0 " << LEGS_BACKWARDS_SHIFT << "\n";
  *t << "  rotation 1 0 0 " << (left ? 0.0 : PI) << "\n";
  *t << "  children [" << "\n";
  *t << "    DEF GROUP_FEMUR_"<< leg_number <<" Group {" << "\n";
  *t << "     children [" << "\n";
  *t << "        DEF TRANSFORM_HIP_"<< leg_number <<" Transform {" << "\n";
  *t << "          translation "<< (left ? -0.004 : 0.004) << " 0 0" << "\n";
  *t << "          rotation 0 0 1 "<< PI_2 << "\n";
  *t << "          children [" << "\n";

  if (leg_number == 1) {
    *t << "            DEF SHAPE_HIP Shape {" << "\n";
    *t << "              appearance DEF LEG_COLOR Appearance {" << "\n";
    *t << "                material Material {" << "\n";
    *t << "                  emissiveColor 0 0 0" << "\n";
    *t << "                }" << "\n";
    *t << "              }" << "\n";
    *t << "              geometry Cylinder {" << "\n";
    *t << "                height " << HIP_LENGTH << "\n";
    *t << "                radius " << HIP_RADIUS << "\n";
    *t << "              }" << "\n";
    *t << "            }" << "\n";
  } else
    *t << "            USE SHAPE_HIP" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "        DEF TRANSFORM_FEMUR_"<< leg_number <<" Transform {" << "\n";
  *t << "          translation " << legs[leg_index].femur.transl_x + HIP_LENGTH / (left ? -2.0 : 2.0) <<" 0 0" << "\n";
  *t << "          rotation 0 0 1 " << PI_2  << "\n";
  *t << "          children [" << "\n";
  *t << "            DEF FEMUR Shape {" << "\n";
  *t << "              appearance USE LEG_COLOR" << "\n";
  *t << "              geometry Cylinder {" << "\n";
  *t << "                height " << femur_length - HIP_LENGTH  << "\n";
  *t << "                radius " << femur_radius << "\n";
  *t << "              }" << "\n";
  *t << "            }" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "        DEF TRANSFORM_KNEE_"<< leg_number <<" Transform {" << "\n";
  *t << "          translation " << legs[leg_index].knee.transl_x << " 0 0" << "\n";
  *t << "          children [" << "\n";
  *t << "            DEF KNEE Shape {" << "\n";
  *t << "              appearance USE LEG_COLOR" << "\n";
  *t << "              geometry Sphere {" << "\n";
  *t << "                radius " << femur_radius << "\n";
  *t << "              }" << "\n";
  *t << "            }" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "      ]" << "\n";
  *t << "    }" << "\n";

  *t << "        DEF GROUP_TIBIA_" << leg_number <<" Group {" << "\n";
  *t << "          children [" << "\n";
  *t << "            DEF TRANSFORM_TIBIA_"<< leg_number << " Transform {" << "\n";
  
  // Rigid Tibias Hack
  legs[leg_index].tibia.transl_x += femur_length / (left ? -2.0 : 2.0);
  
  *t << "              translation " << legs[leg_index].tibia.transl_x << " " << (left ? -0.015 : 0.015 ) <<" 0" << "\n";
  *t << "              rotation 0 0 1 " << 2.5 << "\n";
  *t << "              children [" << "\n";
  *t << "                DEF TIBIA Shape {" << "\n";
  *t << "                  appearance USE LEG_COLOR" << "\n";
  *t << "                  geometry Cylinder {" << "\n";
  *t << "                    height " << tibia_length << "\n";
  *t << "                    radius " << tibia_radius << "\n";
  *t << "                  }" << "\n";
  *t << "                }" << "\n";
  *t << "              ]" << "\n";
  *t << "            }" << "\n";
  *t << "            DEF TRANSFORM_FOOT_"<< leg_number << " Transform {" << "\n";
  
  // Rigid Tibias Foot Hack
  legs[leg_index].foot.transl_y = (left ? -0.032 : 0.032);
  
  *t << "              translation " << legs[leg_index].foot.transl_x << " " << legs[leg_index].foot.transl_y << " " << legs[leg_index].foot.transl_z << "\n";
  *t << "              rotation 0 0 1 " << 2*PI << "\n";
  *t << "              children [" << "\n";
  *t << "                DEF FOOT Shape {" << "\n";
  *t << "                  appearance USE LEG_COLOR" << "\n";
  *t << "                  geometry Cylinder {" << "\n";
  *t << "                    height " << foot_height << "\n";
  *t << "                    radius " << foot_radius << "\n";
  *t << "                  }" << "\n";
  *t << "                }" << "\n";
  *t << "              ]" << "\n";
  *t << "            }" << "\n";
  *t << "          ]" << "\n";
  *t << "    }" << "\n";
  *t << "  ]" << "\n";
  *t << "  name \"servo_leg_"<< leg_number << "\"" << "\n";
  *t << "  boundingObject USE TRANSFORM_FOOT_" << leg_number << "\n";
  *t << "    physics Physics {" << "\n";
  *t << "      density " << leg_density << "\n";
  *t << "      bounce " << BOUNCE_PARAM << "\n";
  *t << "      coulombFriction " << coulomb_friction << "\n";
  *t << "      forceDependentSlip " << force_dependent_slip << "\n";
  *t << "      centerOfMass " << femur_length / (left ? -2.0 : 2.0) << " 0 0" << "\n";
  *t << "    }" << "\n";
  *t << "  joint Joint {" << "\n";
  *t << "  }" << "\n";
  *t << "  maxVelocity " << max_leg_velocity << "\n";
  *t << "  maxForce " << max_force << "\n";
  *t << "}" << "\n";
}


/************************************************************************************
 *
 * 				   MISCELANIA
 *
 ***********************************************************************************/

/*
 * Writes a Fly on the scenario
 */
void CentipedeScript::writeFly( QTextStream *t )
{
  *t << "DEF FLY CustomRobot {" << "\n";
  *t << "  translation 0 0.3 0" << "\n";
  *t << "  children [" << "\n";
  *t << "    DEF FLY_SHAPE Shape {" << "\n";
  *t << "      appearance Appearance {" << "\n";
  *t << "        material Material {" << "\n";
  *t << "          diffuseColor 0 0 0" << "\n";
  *t << "        }" << "\n";
  *t << "      }" << "\n";
  *t << "      geometry Sphere {" << "\n";
  *t << "        radius 0.04" << "\n";
  *t << "      }" << "\n";
  *t << "    }" << "\n";
  *t << "  ]" << "\n";
  *t << "  boundingObject USE FLY_SHAPE" << "\n";
  *t << "  controller \"fly\"" << "\n";
  *t << "}" << "\n";
}

int CentipedeScript::isLeg( QString s )
{
	if(s=="B") return 1;
	if(s=="E") return 1;
	return 0;
}

/*
 * Opens the origin file and writes its content to the destination stream
 */
int CentipedeScript::writeToWorld( QString origin, QTextStream *destination )
{
	if(origin.isEmpty()) return FILE_NO_EXISTS;
	
	char buffer[256];
	
	// Open the file
	std::ifstream file(origin, std::ios::in); 
	// Read the file
	while (file.good()) {
		file.getline(buffer,256);
		*destination << buffer << "\n";
	}
	file.close();	

	return 0;
}


/************************************************************************************
 *
 * 				SCENENARIO WEBOTS CODE
 *
 ***********************************************************************************/

/*
 * Writes the header of the Webots World. Adapted from Yvan
 */
void CentipedeScript::writeWorld( QTextStream *t )
{
  *t << "#VRML_SIM V4.0 utf8" << "\n";
  *t << "#000000" << "\n";
  *t << "WorldInfo {" << "\n";
  *t << "  info [" << "\n";
  *t << "    \"Centipede\"" << "\n";
  *t << "  ]" << "\n";
  *t << "  title \"Centipede\"" << "\n";
  *t << ((physics)? "  physics \"centipede_physics\"" : "") << "\n";
  *t << "  basicTimeStep 4" << "\n";
  *t << "}" << "\n";
  *t << "Viewpoint {" << "\n";
  *t << "  orientation -0.0615918 0.90633 0.418058 2.81691" << "\n";
  *t << "  position 2.81264 7.41319 -6.61422" << "\n";
  *t << "}" << "\n";
  *t << "Background {" << "\n";
  *t << "  skyColor [" << "\n";
  *t << "    1.0 1.0 1.0" << "\n";
  *t << "  ]" << "\n";
  *t << "}" << "\n";
  *t << "DirectionalLight {" << "\n";
  *t << "  direction 0.5 -0.7 -0.2" << "\n";
  *t << "  intensity 0.7" << "\n";
  *t << "}" << "\n";
  *t << "DirectionalLight {" << "\n";
  *t << "  direction -0.5 0.7 0.2" << "\n";
  *t << "  intensity 0.7" << "\n";
  *t << "}" << "\n";
}


/*
 * Writes a ground with a slope. Adapted from Yvan. 
 */
void CentipedeScript::writeGround( QTextStream *t ) {
  
  double l = aquarium_length;
  double w = aquarium_width;
  double b = bottom_level;
  double s = shore_level;

  double h, a;
  if (slope_angle != 0.0) {
    h = (s - b) / sin(slope_angle);
    a = (s - b) / tan(slope_angle);
  }
  else
    h = a = l/3;

  *t << "DEF GROUND Solid {" << "\n";
  *t << "  children [" << "\n";
  *t << "    DEF GND Group {" << "\n";
  *t << "      children [" << "\n";
  *t << "        Transform {" << "\n";
  *t << "          translation " << -(a/2 + (l-a)/4) << " " << b << " 0" << "\n";
  *t << "          children [" << "\n";
  *t << "            Shape {" << "\n";
  *t << "              appearance DEF GROUND_APPEARANCE Appearance {"  << "\n";
  *t << "                material Material {" << "\n";
  *t << "                  diffuseColor 0.8 0.8 0.8" << "\n";
  *t << "                  ambientIntensity 0" << "\n";
  *t << "                  transparency 0" << "\n";
  *t << "                }" << "\n";
  *t << "              }" << "\n";
  *t << "              geometry Box {" << "\n";
  *t << "                size " << (l-a)/2 << " 0.001 " << w << "\n";
  *t << "              }" << "\n";
  *t << "            }" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "        Transform {" << "\n";
  *t << "          translation 0 " << b+(s-b)/2 << " 0" << "\n";
  *t << "          rotation 0 0 1 " << slope_angle << "\n";
  *t << "          children [" << "\n";
  *t << "            Shape {" << "\n";
  *t << "              appearance USE GROUND_APPEARANCE" << "\n";
  *t << "              geometry Box {" << "\n";
  *t << "                size " << h << " 0.001 " << w << "\n";
  *t << "              }" << "\n";
  *t << "            }" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "        Transform {" << "\n";
  *t << "          translation " << a/2+(l-a)/4 << " " << s << " 0" << "\n";
  *t << "          children [" << "\n";
  *t << "            Shape {" << "\n";
  *t << "              appearance USE GROUND_APPEARANCE" << "\n";
  *t << "              geometry Box {" << "\n";
  *t << "                size " << (l-a)/2 << " 0.001 " << w << "\n";
  *t << "              }" << "\n";
  *t << "            }" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "      ]" << "\n";
  *t << "    }" << "\n";
  *t << "  ]" << "\n";
  *t << "  name \"ground\"" << "\n";
  *t << "  boundingObject USE GND" << "\n";
  *t << "  locked TRUE" << "\n";
  *t << "}" << "\n";
}

void CentipedeScript::writeGridGround( QTextStream *t ) {

  *t << "DEF GROUND Solid {" << "\n";
  *t << "  children [" << "\n";
  *t << "    Shape {" << "\n";
  *t << "      appearance Appearance {" << "\n";
  *t << "        material Material {" << "\n";
  *t << "          diffuseColor 0.301961 0.301961 0.301961" << "\n";
  *t << "        }" << "\n";
  *t << "        texture ImageTexture {" << "\n";
  *t << "          url [" << "\n";
  *t << "            \"textures/texture_grid.png\"" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "      }" << "\n";
  *t << "      geometry DEF GROUND_FACE_SET IndexedFaceSet {" << "\n";
  *t << "        coord Coordinate {" << "\n";
  *t << "          point [" << "\n";
  *t << "            50 0 50" << "\n";
  *t << "            -50 0 50" << "\n";
  *t << "            -50 0 -50" << "\n";
  *t << "            50 0 -50" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "        texCoord TextureCoordinate {" << "\n";
  *t << "          point [" << "\n";
  *t << "            12 12" << "\n";
  *t << "            12 -12" << "\n";
  *t << "            -12 -12" << "\n";
  *t << "            -12 12" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "        coordIndex [" << "\n";
  *t << "          3, 2, 1, 0" << "\n";
  *t << "        ]" << "\n";
  *t << "        texCoordIndex [" << "\n";
  *t << "          0, 1, 2, 3" << "\n";
  *t << "        ]" << "\n";
  *t << "      }" << "\n";
  *t << "    }" << "\n";
  *t << "    Shape {" << "\n";
  *t << "      appearance Appearance {" << "\n";
  *t << "        material Material {" << "\n";
  *t << "          diffuseColor 0 0 0" << "\n";
  *t << "        }" << "\n";
  *t << "      }" << "\n";
  *t << "      geometry Box {" << "\n";
  *t << "        size 0.5 0.01 0.5" << "\n";
  *t << "      }" << "\n";
  *t << "    }" << "\n";
  *t << "  ]" << "\n";
  *t << "  boundingObject Transform {" << "\n";
  *t << "    translation 0 -1 0" << "\n";
  *t << "    children [" << "\n";
  *t << "      Box {" << "\n";
  *t << "        size 100 2 100" << "\n";
  *t << "      }" << "\n";
  *t << "    ]" << "\n";
  *t << "  }" << "\n";
  *t << "  locked TRUE" << "\n";
  *t << "}" << "\n";
}

/*
 * Writes an infinite flat ground
 */
void CentipedeScript::writeInfiniteGround( QTextStream *t ) {

	*t << "DEF GROUND Solid {" << "\n";
	*t << "  children [" << "\n";
	*t << "    DEF GND Group {" << "\n";
	*t << "      children [" << "\n";
	*t << "        Transform {" << "\n";
	*t << "          children [" << "\n";
	*t << "            Shape {" << "\n";
	*t << "              appearance DEF GROUND_APPEARANCE Appearance {" << "\n";
	*t << "                material Material {" << "\n";
	*t << "                  ambientIntensity 0" << "\n";
	*t << "                }" << "\n";
	*t << "              }" << "\n";
	*t << "              geometry Box {" << "\n";
	*t << "                size 9999 0.1 9999" << "\n";
	*t << "              }" << "\n";
	*t << "            }" << "\n";
	*t << "          ]" << "\n";
	*t << "        }" << "\n";
	*t << "      ]" << "\n";
	*t << "    }" << "\n";
	*t << "  ]" << "\n";
	*t << "  name \"ground\"" << "\n";
	*t << "  boundingObject USE GND" << "\n";
	*t << "  locked TRUE" << "\n";
	*t << "}" << "\n";
}

/*
 * Writes an infinite water scenario
 */
void CentipedeScript::writeInfiniteWater( QTextStream *t ) {
	*t << "DEF WATER Solid {" << "\n";
	*t << "  translation 0 -0.125 0" << "\n";
	*t << "  children [" << "\n";
	*t << "    Shape {" << "\n";
	*t << "      appearance Appearance {" << "\n";
	*t << "        material Material {" << "\n";
	*t << "          diffuseColor 0 0.498039 1" << "\n";
	*t << "          emissiveColor 0 0 1" << "\n";
	*t << "          transparency 0.7" << "\n";
	*t << "        }" << "\n";
	*t << "      }" << "\n";
	*t << "      geometry Box {" << "\n";
	*t << "        size 9999 0.25 9999" << "\n";
	*t << "      }" << "\n";
	*t << "    }" << "\n";
	*t << "  ]" << "\n";
	*t << "  locked TRUE" << "\n";
	*t << "}" << "\n";
}

/*
 * Writes an aquarium. Adapted from Yvan.
 */
void CentipedeScript::writeAquarium( QTextStream *t ) {

  double l = aquarium_length;
  double w = aquarium_width;
  double s = water_level > shore_level ? water_level : shore_level;
  double b = water_level < bottom_level ? water_level : bottom_level;

  *t << "DEF AQUARIUM Solid {" << "\n";
  *t << "  translation 0 " << -(s-b)/2 + s + 0.1 << " 0" << "\n";
  *t << "  children [" << "\n";
  *t << "    DEF SIDES_GROUP Group {" << "\n";
  *t << "      children [" << "\n";
  *t << "        Transform {" << "\n";
  *t << "          translation 0 0 " << w / 2 << "\n";
  *t << "          children [" << "\n";
  *t << "            DEF GLASS_SHAPE1 Shape {" << "\n";
  *t << "              appearance DEF GLASS_APPEARANCE Appearance {" << "\n";
  *t << "                material Material {" << "\n";
  *t << "                  diffuseColor 1 0.6 0.6" << "\n";
  *t << "                }" << "\n";
  *t << "              }" << "\n";
  *t << "              geometry Box {" << "\n";
  *t << "                size " << l << " " << s-b+0.2 << " 0.02" << "\n";
  *t << "              }" << "\n";
  *t << "            }" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "        Transform {" << "\n";
  *t << "          translation 0 0 " << -w / 2 << "\n";
  *t << "          children [" << "\n";
  *t << "            USE GLASS_SHAPE1" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "        Transform {" << "\n";
  *t << "          translation " << -l/2 << " 0 0" << "\n";
  *t << "          rotation 0 1 0 1.5708" << "\n";
  *t << "          children [" << "\n";
  *t << "            DEF GLASS_SHAPE2 Shape {" << "\n";
  *t << "              appearance USE GLASS_APPEARANCE" << "\n";
  *t << "              geometry Box {" << "\n";
  *t << "                size " << w << " " << s - b + 0.2 << " 0.02" << "\n";
  *t << "              }" << "\n";
  *t << "            }" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "        Transform {" << "\n";
  *t << "          translation " << l/2 << " 0 0" << "\n";
  *t << "          rotation 0 1 0 1.5708" << "\n";
  *t << "          children [" << "\n";
  *t << "            USE GLASS_SHAPE2" << "\n";
  *t << "          ]" << "\n";
  *t << "        }" << "\n";
  *t << "      ]" << "\n";
  *t << "    }" << "\n";
  *t << "  ]" << "\n";
  *t << "  boundingObject USE SIDES_GROUP" << "\n";
  *t << "  locked TRUE" << "\n";
  *t << "}" << "\n";
}

/*
 * Writes the aquarium's water. Adapted from Yvan.
 */
void CentipedeScript::writeWater( QTextStream *t ) 
{
  double l = aquarium_length;
  double w = aquarium_width;
  double a = water_level - bottom_level;

  *t << "DEF WATER Solid {" << "\n";
  *t << "  translation 0 " << (water_level - a/2) <<" 0" << "\n";
  *t << "  children [" << "\n";
  *t << "    Shape {" << "\n";
  *t << "      appearance Appearance {" << "\n";
  *t << "        material Material {" << "\n";
  *t << "          diffuseColor 0 0.498039 1" << "\n";
  *t << "          emissiveColor 0 0 1" << "\n";
  *t << "          transparency 0.7" << "\n";
  *t << "        }" << "\n";
  *t << "      }" << "\n";
  *t << "      geometry Box {" << "\n";
  *t << "        size " << l << " " << a << " " << w << "\n";
  *t << "      }" << "\n";
  *t << "    }" << "\n";
  *t << "  ]" << "\n";
  *t << "  locked TRUE" << "\n";
  *t << "}" << "\n";
}
