/*----------------------------------------------------------------------------
 * defines.h
 *
 * Author:  Jerome Braure <jerome.braure@epfl.ch>
 *          Adapted by Yvan Bourquin
 *			Adapted by Brian Jimenez
 * Date:    2007-04-19
 * Release: 1.0
 * Purpose: Contains all the '#define's that are used by different files
 *----------------------------------------------------------------------------
 */

#include <math.h>
 
#ifndef DEFINES_H
#define DEFINES_H

/* general things */
#define TRUE             1
#define FALSE            0
#define MAX_BUF          300
#define SUCCESS          1
#define FAILURE          -1
#define SPACE            ' '
#define PI               3.141592654
#define PI_2             1.570796327        /* PI/2 */
#define X                0
#define Y                1
#define Z                2

#define FILE_ERROR	-5
#define FILE_NO_EXISTS	-4
#define WRITING_WORLD_ERROR	-3
#define GENERAL_ERROR	-1

/* world settings */
#define WATER_DENSITY  1000        /* [kg/m³] */
#define GRAVITY        9.81        /* [m/s²] */

/* experimental settings, AC 8.9.2004 */
#define SWIMMING_EXPERIMENTAL_PROCEDURE    0

/* replace inertial forces with viscosity */
#define ENABLE_VISCOSITY     0
#define WATER_VISCOSITY      1.025e-3      /* Pa.s @ 20 °C */
#define VISCOSITY_K          16            /* 16 = disc .. 6 * PI = sphere */

/* salamander fixed settings */
#define SEGMENT_HEIGHT       0.056        /* [m] */
#define SEGMENT_LENGTH       0.095        /* [m] */
#define SEGMENT_WIDTH        0.038        /* [m] */
#define SEGMENT_WIDTH_HALF   (SEGMENT_WIDTH/2.0)        /* [m] */
#define SEGMENT_CYL_RADIUS   SEGMENT_WIDTH_HALF
#define SEGMENT_CYL_VOLUME   (PI*SQUARE(SEGMENT_CYL_RADIUS)*SEGMENT_HEIGHT)
#define SEGMENT_BOX_VOLUME   (SEGMENT_LENGTH*SEGMENT_WIDTH*SEGMENT_HEIGHT)
#define SEGMENT_TOTAL_VOLUME (SEGMENT_CYL_VOLUME + SEGMENT_BOX_VOLUME)
#define SEGMENT_CX           2.566666        /* depends on segment geometry */
#define SEGMENT_CY           1.95        /* depends on segment geometry */

/* added for new amphibot design by yvan.bourquin@epfl.ch */
#define BIG_SEGMENT_LENGTH     0.079
#define BIG_SEGMENT_WIDTH      0.064
#define BIG_SEGMENT_WIDTH_HALF (BIG_SEGMENT_WIDTH/2.0)
#define LEGS_BACKWARDS_SHIFT   0.02
#define HIP_LENGTH             0.008
#define HIP_RADIUS             0.00975

#define BODY_FREQUENCE       1
#define SERVO_VELOCITY       1000000        /* [rad/s], large for better trajectory following */
#define FRICTION_MU1         0.5        /* not used, physics reads file to get it. */
#define FRICTION_MU2         0.1        /* not used anymore, AC 31.8.2004 */
#define BOUNCE_PARAM         0.5
/* drag coefficients (lambda = 0.5 * rho_water * Surface_perp * C_drag) */
#define LAMBDA_X (0.5 * WATER_DENSITY * SEGMENT_LENGTH \
                  * SEGMENT_HEIGHT * SEGMENT_CX)
#define LAMBDA_Y (0.5 * WATER_DENSITY * SEGMENT_LENGTH \
                  * SEGMENT_WIDTH * SEGMENT_CY)

/* simulation settings */
/* [#define DELTA_T                0.02        s] */
#define DELTA_T                0.005        /* [s] */
#define STEP                   (DELTA_T * 1000)        /* simulation time step [ms] */
#define LOCOMOTION_SNAKE       0
#define LOCOMOTION_WALK        1
#define POSITION_IN_WATER      1
#define POSITION_AT_SURFACE    2
#define POSITION_OUT_OF_WATER  3
#define MAX_TIME               30        /* time before simulation is quitting [s] */
#define NBR_OF_TIMESTEPS       (MAX_TIME/DELTA_T)        /* total nbr of steps in a run */
#define WINDOW_SIZE            50        /* nbr of samples for average speed */
#define GRAPH_TIME             500        /* x-axis length in gnuplot (history length) */
#define STEADY_AVG_TOLERANCE   0.05        /* 5% */

/* macros */
#define NORM2(x, y)                 (sqrt(SQUARE(x) + SQUARE(y)))
#define NORM3(x, y, z)              (sqrt(SQUARE(x) + SQUARE(y) + SQUARE(z)))
#define MIN(x, y)                   (((x)>(y))?(y):(x))
#define MAX(x, y)                   (((x)<(y))?(y):(x))
#define RAD_TO_DEG(rad)             (180*(rad)/PI)
#define DEG_TO_RAD(deg)             (PI*(deg)/180)
#define SQUARE(x)                   ((x)*(x))
#define CUBE(x)                     ((x)*(x)*(x))
#define SPHERE_VOLUME(radius)       ((4/3) * PI * CUBE(radius))
#define CYLINDER_VOLUME(r, h)       (PI * SQUARE(r) * (h))
#define VECTOR_PRODUCT(a, b, c)     {c[0] = a[1]*b[2] - a[2]*b[1]; \
                                     c[1] = a[2]*b[0] - a[0]*b[2]; \
                                     c[2] = a[0]*b[1] - a[1]*b[0];}
#define SCALAR_PRODUCT(a, b)        (a[0]*b[0]+a[1]*b[1]+a[2]*b[2])

#define MATRIX_PRINT(A){ \
   printf(#A " = \n"); \
   printf(" | %.2f  %.2f %.2f |\n",   A[0][0], A[0][1], A[0][2]); \
   printf(" | %.2f  %.2f %.2f |\n",   A[1][0], A[1][1], A[1][2]); \
   printf(" | %.2f  %.2f %.2f |\n\n", A[2][0], A[2][1], A[2][2]); \
}

#define MATRIX_TRANSPOSE(A, AT){ \
  int i = 0; \
  int j; \
  for(; i<3; i++){ \
    for(j = 0; j<3; j++){ \
      AT[i][j] = A[j][i]; \
    } \
  } \
}

/* A * B = C 
 * A, B, C must be 3x3.
 * A[line][column]
 */
#define MATRIX_MULTIPLY(A, B, C){ \
  int i = 0; \
  int j = 0; \
  for(; i<3; i++){ \
    for(j = 0; j<3; j++){ \
      C[i][j] = A[i][0]*B[0][j] + A[i][1]*B[1][j] \
              + A[i][2]*B[2][j]; \
    } \
  } \
}

/* A * b = c
 * A must be 3x3.
 * b and c must be 3x1
 * A[line][column]
 */
#define MATRIX_VECTOR_MULTIPLY(A, b, c){ \
  c[0] = A[0][0]*b[0] + A[0][1]*b[1] + A[0][2]*b[2]; \
  c[1] = A[1][0]*b[0] + A[1][1]*b[1] + A[1][2]*b[2]; \
  c[2] = A[2][0]*b[0] + A[2][1]*b[1] + A[2][2]*b[2]; \
}
#define DEBUG(val)  printf(#val " = %f\n", val)
#define NORM(vect)  (sqrt(SQUARE(vect[0])+SQUARE(vect[1])+SQUARE(vect[2])))

struct s_coord {
  float x, y, z;
};

#endif
